@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2022 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
+   CA 94945, U.S.A., +1(415)492-9861, for further information.
 */
 
 
@@ -25,6 +25,7 @@
 #include "strimpl.h"
 #include "sa85x.h"
 #include "sarc4.h"
+#include "sstring.h"
 
 #define CHECK(expr)\
   BEGIN if ((code = (expr)) < 0) return code; END
@@ -168,7 +169,7 @@ cos_object_init(cos_object_t *pco, gx_device_pdf *pdev,
         pco->id = 0;
         pco->elements = 0;
         pco->pieces = 0;
-        pco->pdev = pdev;
+        pco->mem = pdev->pdf_memory;
         pco->pres = 0;
         pco->is_open = true;
         pco->is_graphics = false;
@@ -185,7 +186,7 @@ cos_object_init(cos_object_t *pco, gx_device_pdf *pdev,
 gs_memory_t *
 cos_object_memory(const cos_object_t *pco)
 {
-    return pco->pdev->pdf_memory;
+    return pco->mem;
 }
 
 /* Change a generic cos object into one of a specific type. */
@@ -280,12 +281,12 @@ cos_resource_value(cos_value_t *pcv, cos_object_t *pco)
 
 /* Free a value. */
 void
-cos_value_free(const cos_value_t *pcv, const cos_object_t *pco,
+cos_value_free(const cos_value_t *pcv, gs_memory_t *mem,
                client_name_t cname)
 {
     switch (pcv->value_type) {
     case COS_VALUE_SCALAR:
-        gs_free_string(cos_object_memory(pco), pcv->contents.chars.data,
+        gs_free_string(mem, pcv->contents.chars.data,
                        pcv->contents.chars.size, cname);
     case COS_VALUE_CONST:
         break;
@@ -510,15 +511,14 @@ cos_array_from_floats(gx_device_pdf *pdev, const float *pf, uint size,
 static void
 cos_array_release(cos_object_t *pco, client_name_t cname)
 {
-    gs_memory_t *mem = cos_object_memory(pco);
     cos_array_t *const pca = (cos_array_t *)pco;
     cos_array_element_t *cur;
     cos_array_element_t *next;
 
     for (cur = pca->elements; cur; cur = next) {
         next = cur->next;
-        cos_value_free(&cur->value, pco, cname);
-        gs_free_object(mem, cur, cname);
+        cos_value_free(&cur->value, cos_object_memory(pco), cname);
+        gs_free_object(cos_object_memory(pco), cur, cname);
     }
     pca->elements = 0;
 }
@@ -558,7 +558,7 @@ cos_array_write(const cos_object_t *pco, gx_device_pdf *pdev, gs_id object_id)
                 /* Careful here, only certain errors will bubble up
                  * through the text processing.
                  */
-                return gs_error_limitcheck;
+                return_error(gs_error_limitcheck);
                 break;
             default:
                 emprintf(pdev->memory,
@@ -654,7 +654,7 @@ cos_array_put_no_copy(cos_array_t *pca, long index, const cos_value_t *pvalue)
         ppcae = &next->next;
     if (next && next->index == index) {
         /* We're replacing an existing element. */
-        cos_value_free(&next->value, COS_OBJECT(pca),
+        cos_value_free(&next->value, mem,
                        "cos_array_put(old value)");
         pcae = next;
     } else {
@@ -701,7 +701,7 @@ cos_array_add_int(cos_array_t *pca, int i)
     char str[sizeof(int) * 8 / 3 + 3]; /* sign, rounding, 0 terminator */
     cos_value_t v;
 
-    gs_sprintf(str, "%d", i);
+    gs_snprintf(str, sizeof(str), "%d", i);
     return cos_array_add(pca, cos_string_value(&v, (byte *)str, strlen(str)));
 }
 int
@@ -807,7 +807,7 @@ cos_dict_element_free(cos_dict_t *pcd, cos_dict_element_t *pcde,
 {
     gs_memory_t *mem = COS_OBJECT_MEMORY(pcd);
 
-    cos_value_free(&pcde->value, COS_OBJECT(pcd), cname);
+    cos_value_free(&pcde->value, mem, cname);
     if (pcde->owns_key)
         gs_free_string(mem, pcde->key.data, pcde->key.size, cname);
     gs_free_object(mem, pcde, cname);
@@ -891,7 +891,7 @@ cos_elements_write(stream *s, const cos_dict_element_t *pcde,
                         /* Careful here, only certain errors will bubble up
                          * through the text processing.
                          */
-                        return gs_error_limitcheck;
+                        return_error(gs_error_limitcheck);
                         break;
                     default:
                         emprintf(pdev->memory,
@@ -933,69 +933,6 @@ cos_dict_write(const cos_object_t *pco, gx_device_pdf *pdev, gs_id object_id)
     return 0;
 }
 
-static int find_last_dict_entry(const cos_dict_t *d, const cos_dict_element_t **element)
-{
-    const cos_dict_element_t *pcde = d->elements, *Last;
-    int code, length, length1, length2, offset1 = 0, offset2 = 0, i;
-
-    *element = 0L;
-
-    Last = pcde;
-    for (i = 0;Last->key.data[i] == 0x00; i++)
-        ;
-    length1 = Last->key.size - i;
-    offset1 = i;
-
-    if (Last->key.data[offset1] == '/') {
-        length1 -= 1;
-        offset1 += 1;
-    } else {
-        if (pcde->key.data[offset1] == '(') {
-            length1 -= 2;
-            offset1 = 1;
-        } else {
-            return_error(gs_error_typecheck);
-        }
-    }
-
-    pcde = pcde->next;
-    while (pcde){
-        for (i = 0;pcde->key.data[i] == 0x00; i++)
-            ;
-        length2 = pcde->key.size - i;
-        offset2 = i;
-        if (pcde->key.data[offset2] == '/') {
-            length2 -= 1;
-            offset2 += 1;
-        } else {
-            if (pcde->key.data[offset2] == '(') {
-                length2 -= 2;
-                offset2 = 1;
-            } else {
-                return_error(gs_error_typecheck);
-            }
-        }
-
-        if (length2 < length1)
-            length = length2;
-        else
-            length = length1;
-        code = strncmp((const char *)&pcde->key.data[offset2], (const char *)&Last->key.data[offset1], length);
-        if (code == 0) {
-            if (length2 > length1) {
-                Last = pcde;
-                length1 = length2;
-            }
-        } else if (code > 0) {
-            Last = pcde;
-            length1 = length2;
-        }
-        pcde = pcde->next;
-    }
-    *element = Last;
-    return 0;
-}
-
 static int find_first_dict_entry(const cos_dict_t *d, const cos_dict_element_t **element)
 {
     const cos_dict_element_t *pcde = d->elements, *First;
@@ -1055,10 +992,12 @@ static int find_first_dict_entry(const cos_dict_t *d, const cos_dict_element_t *
             if (length2 < length1) {
                 First = pcde;
                 length1 = length2;
+                offset1 = offset2;
             }
         } else if (code < 0) {
             First = pcde;
             length1 = length2;
+            offset1 = offset2;
         }
         pcde = pcde->next;
     }
@@ -1122,21 +1061,8 @@ static int find_next_dict_entry(const cos_dict_t *d, const cos_dict_element_t **
                 code = strncmp((const char *)&pcde->key.data[offset2], (const char *)&Next->key.data[offset3], length);
                 if (code < 0 || (code == 0 && length3 > length2)) {
                     Next = pcde;
-                    for (i = 0;Next->key.data[i] == 0x00; i++)
-                        ;
-                    length3 = Next->key.size - i;
-                    offset3 = i;
-                    if (Next->key.data[offset3] == '/') {
-                        length3 -= 1;
-                        offset3 += 1;
-                    } else {
-                        if (Next->key.data[0] == '(') {
-                            length3 -= 2;
-                            offset3 = 1;
-                        } else {
-                            return_error(gs_error_typecheck);
-                        }
-                    }
+                    length3 = length2;
+                    offset3 = offset2;
                 }
             } else {
                 Next = pcde;
@@ -1146,11 +1072,11 @@ static int find_next_dict_entry(const cos_dict_t *d, const cos_dict_element_t **
                 offset3 = i;
                 if (Next->key.data[offset3] == '/') {
                     length3 -= 1;
-                    offset3 = 1;
+                    offset3 += 1;
                 } else {
                     if (Next->key.data[0] == '(') {
                         length3 -= 2;
-                        offset3 = 1;
+                        offset3 += 1;
                     } else {
                         return_error(gs_error_typecheck);
                     }
@@ -1163,7 +1089,59 @@ static int find_next_dict_entry(const cos_dict_t *d, const cos_dict_element_t **
     return 0;
 }
 
-static int write_key_as_string(stream *s, const cos_dict_element_t *element)
+static int find_last_dict_entry(const cos_dict_t *d, const cos_dict_element_t **element)
+{
+    const cos_dict_element_t *pcde = d->elements, *Last, *Next;
+
+    *element = 0L;
+
+    Next = Last = pcde;
+
+    do {
+        Last = Next;
+        find_next_dict_entry(d, &Next);
+    } while (Next);
+
+    *element = Last;
+
+    return 0;
+}
+static int write_key_as_string_encrypted(const gx_device_pdf *pdev, const byte *str, uint size, gs_id object_id)
+{
+    stream sout;
+    stream_PSSD_state st;
+    stream_state so;
+    byte bufo[100], *buffer;
+    stream_arcfour_state sarc4;
+
+    buffer = gs_alloc_bytes(pdev->pdf_memory, size, "encryption buffer");
+    if (buffer == 0L)
+        return 0;
+
+    if (pdf_encrypt_init(pdev, object_id, &sarc4) < 0) {
+        gs_free_object(pdev->pdf_memory, buffer, "Free encryption buffer");
+        /* The interface can't pass an error. */
+        stream_write(pdev->strm, str, size);
+        return size;
+    }
+    s_init_state((stream_state *)&st, &s_PSSD_template, NULL);
+    s_init(&sout, NULL);
+    s_init_state(&so, &s_PSSE_template, NULL);
+    s_init_filter(&sout, &so, bufo, sizeof(bufo), pdev->strm);
+    stream_putc(pdev->strm, '(');
+    memcpy(buffer, str, size);
+    s_arcfour_process_buffer(&sarc4, buffer, size);
+    stream_write(&sout, buffer, size);
+    /* Another case where we use sclose() and not s_close_filters(), because the
+     * buffer we supplied to s_init_filter is a heap based C object, so we
+     * must not free it.
+     */
+    sclose(&sout); /* Writes ')'. */
+    gs_free_object(pdev->pdf_memory, buffer, "Free encryption buffer");
+    return 0;
+}
+
+static int write_key_as_string(const gx_device_pdf *pdev, stream *s, const cos_dict_element_t *element, gs_id object_id)
 {
     int i, length, offset;
 
@@ -1184,11 +1162,18 @@ static int write_key_as_string(stream *s, const cos_dict_element_t *element)
     if (element->key.data[offset] == '/') {
         offset++;
         length--;
-        spputc(s, '(');
-        stream_write(s, &element->key.data[offset], length);
-        spputc(s, ')');
+        if (!pdev->KeyLength || object_id == (gs_id)-1) {
+            spputc(s, '(');
+            stream_write(s, &element->key.data[offset], length);
+            spputc(s, ')');
+        }
+        else
+            write_key_as_string_encrypted(pdev, &element->key.data[offset], length, object_id);
     } else {
-        stream_write(s, element->key.data, element->key.size);
+        if (!pdev->KeyLength || object_id == (gs_id)-1)
+            stream_write(s, element->key.data, element->key.size);
+        else
+            write_key_as_string_encrypted(pdev, &element->key.data[1], element->key.size - 2, object_id);
     }
     return 0;
 }
@@ -1231,14 +1216,14 @@ cos_write_dict_as_ordered_array(cos_object_t *pco, gx_device_pdf *pdev, pdf_reso
     }
 
     stream_puts(s, "<<\n/Limits [\n");
-    write_key_as_string(s, First);
+    write_key_as_string(pdev, s, First, pco->id);
     stream_puts(s, "\n");
-    write_key_as_string(s, Last);
+    write_key_as_string(pdev, s, Last, pco->id);
     stream_puts(s, "\n]\n");
     stream_puts(s, "/Names [");
     do {
         stream_puts(s, "\n");
-        write_key_as_string(s, First);
+        write_key_as_string(pdev, s, First, pco->id);
         cos_value_write_spaced(&First->value, pdev, true, -1);
         find_next_dict_entry(d, &First);
     } while (First);
@@ -1323,10 +1308,7 @@ cos_dict_put_copy(cos_dict_t *pcd, const byte *key_data, uint key_size,
                                       (flags & DICT_COPY_VALUE) != 0);
         if (code < 0)
             return code;
-        if (flags & DICT_FREE_KEY)
-            gs_free_const_string(mem, key_data, key_size,
-                                 "cos_dict_put(new key)");
-        cos_value_free(&next->value, COS_OBJECT(pcd),
+        cos_value_free(&next->value, mem,
                        "cos_dict_put(old value)");
         pcde = next;
     } else {
@@ -1398,7 +1380,7 @@ cos_dict_put_c_key_int(cos_dict_t *pcd, const char *key, int value)
 {
     char str[sizeof(int) * 8 / 3 + 3]; /* sign, rounding, 0 terminator */
 
-    gs_sprintf(str, "%d", value);
+    gs_snprintf(str, sizeof(str), "%d", value);
     return cos_dict_put_c_key_string(pcd, key, (byte *)str, strlen(str));
 }
 int
@@ -1420,10 +1402,10 @@ cos_dict_put_c_key_real(cos_dict_t *pcd, const char *key, double value)
     return cos_dict_put_c_key_string(pcd, key, str, stell(&s));
 }
 int
-cos_dict_put_c_key_floats(cos_dict_t *pcd, const char *key, const float *pf,
+cos_dict_put_c_key_floats(gx_device_pdf *pdev, cos_dict_t *pcd, const char *key, const float *pf,
                           uint size)
 {
-    cos_array_t *pca = cos_array_from_floats(pcd->pdev, pf, size,
+    cos_array_t *pca = cos_array_from_floats(pdev, pf, size,
                                              "cos_dict_put_c_key_floats");
     int code;
 
@@ -1580,7 +1562,7 @@ cos_param_put_typed(gs_param_list * plist, gs_param_name pkey,
 {
     cos_param_list_writer_t *const pclist =
         (cos_param_list_writer_t *)plist;
-    gx_device_pdf *pdev = pclist->pcd->pdev;
+    gx_device_pdf *pdev = pclist->pdev;
     gs_memory_t *mem = pclist->memory;
     cos_value_t value;
     cos_array_t *pca;
@@ -1588,6 +1570,9 @@ cos_param_put_typed(gs_param_list * plist, gs_param_name pkey,
     byte key_chars[100];		/****** ADHOC ******/
     int code;
 
+    while(pdev->child)
+        pdev = (gx_device_pdf *)pdev->child;
+
     if (key_len > sizeof(key_chars) - 1)
         return_error(gs_error_limitcheck);
     switch (pvalue->type) {
@@ -1655,12 +1640,13 @@ cos_param_put_typed(gs_param_list * plist, gs_param_name pkey,
 }
 
 int
-cos_param_list_writer_init(cos_param_list_writer_t *pclist, cos_dict_t *pcd,
+cos_param_list_writer_init(gx_device_pdf *pdev, cos_param_list_writer_t *pclist, cos_dict_t *pcd,
                            int print_ok)
 {
     gs_param_list_init((gs_param_list *)pclist, &cos_param_list_writer_procs,
                        COS_OBJECT_MEMORY(pcd));
     pclist->pcd = pcd;
+    pclist->pdev = pdev;
     pclist->print_ok = print_ok;
     return 0;
 }
@@ -1689,14 +1675,13 @@ cos_stream_alloc(gx_device_pdf *pdev, client_name_t cname)
 static void
 cos_stream_release(cos_object_t *pco, client_name_t cname)
 {
-    gs_memory_t *mem = cos_object_memory(pco);
     cos_stream_t *const pcs = (cos_stream_t *)pco;
     cos_stream_piece_t *cur;
     cos_stream_piece_t *next;
 
     for (cur = pcs->pieces; cur; cur = next) {
         next = cur->next;
-        gs_free_object(mem, cur, cname);
+        gs_free_object(cos_object_memory(pco), cur, cname);
     }
     pcs->pieces = 0;
     cos_dict_release(pco, cname);
@@ -1706,14 +1691,14 @@ static int hash_cos_stream(const cos_object_t *pco0, gs_md5_state_t *md5, gs_md5
 {
     const cos_stream_t *pcs = (const cos_stream_t *)pco0;
     cos_stream_piece_t *pcsp = pcs->pieces;
-    FILE *sfile = pdev->streams.file;
+    gp_file *sfile = pdev->streams.file;
     byte *ptr;
     int64_t position_save;
     int result;
 
     sflush(pdev->strm);
     sflush(pdev->streams.strm);
-    position_save = gp_ftell_64(sfile);
+    position_save = gp_ftell(sfile);
 
     if (!pcsp)
         return -1;
@@ -1725,8 +1710,10 @@ static int hash_cos_stream(const cos_object_t *pco0, gs_md5_state_t *md5, gs_md5
             result = gs_note_error(gs_error_VMerror);
             return result;
         }
-        gp_fseek_64(sfile, pcsp->position, SEEK_SET);
-        if (fread(ptr, 1, pcsp->size, sfile) != pcsp->size) {
+        if (gp_fseek(sfile, pcsp->position, SEEK_SET) != 0)
+	  return_error(gs_error_ioerror);
+
+        if (gp_fread(ptr, 1, pcsp->size, sfile) != pcsp->size) {
             gs_free(pdev->memory, ptr, sizeof (byte), pcsp->size, "hash_cos_stream");
             result = gs_note_error(gs_error_ioerror);
             return result;
@@ -1735,8 +1722,10 @@ static int hash_cos_stream(const cos_object_t *pco0, gs_md5_state_t *md5, gs_md5
         gs_free(pdev->memory, ptr, sizeof (byte), pcsp->size, "hash_cos_stream");
         pcsp = pcsp->next;
     }
-    gp_fseek_64(sfile, position_save, SEEK_SET);
     gs_md5_finish(md5, (gs_md5_byte_t *)hash);
+    if (gp_fseek(sfile, position_save, SEEK_SET) != 0)
+      return_error(gs_error_ioerror);
+
     return 0;
 }
 
@@ -1817,7 +1806,7 @@ cos_stream_contents_write(const cos_stream_t *pcs, gx_device_pdf *pdev)
     cos_stream_piece_t *pcsp;
     cos_stream_piece_t *last;
     cos_stream_piece_t *next;
-    FILE *sfile = pdev->streams.file;
+    gp_file *sfile = pdev->streams.file;
     int64_t end_pos;
     bool same_file = (pdev->sbstack_depth > 0);
     int code;
@@ -1836,13 +1825,19 @@ cos_stream_contents_write(const cos_stream_t *pcs, gx_device_pdf *pdev)
     for (pcsp = pcs->pieces, last = NULL; pcsp; pcsp = next)
         next = pcsp->next, pcsp->next = last, last = pcsp;
     for (pcsp = last, code = 0; pcsp && code >= 0; pcsp = pcsp->next) {
-        if (same_file)
-            pdf_copy_data_safe(s, sfile, pcsp->position, pcsp->size);
-        else {
-            end_pos = gp_ftell_64(sfile);
-            gp_fseek_64(sfile, pcsp->position, SEEK_SET);
-            pdf_copy_data(s, sfile, pcsp->size, ss);
-            gp_fseek_64(sfile, end_pos, SEEK_SET);
+        if (same_file) {
+            code = pdf_copy_data_safe(s, sfile, pcsp->position, pcsp->size);
+            if (code < 0)
+                return code;
+        } else {
+            end_pos = gp_ftell(sfile);
+            if (gp_fseek(sfile, pcsp->position, SEEK_SET) != 0)
+	      return_error(gs_error_ioerror);
+            code = pdf_copy_data(s, sfile, pcsp->size, ss);
+            if (code < 0)
+                return code;
+            if (gp_fseek(sfile, end_pos, SEEK_SET) != 0)
+	      return_error(gs_error_ioerror);
         }
     }
     /* Reverse the elements back. */
@@ -1887,13 +1882,21 @@ cos_stream_dict(cos_stream_t *pcs)
 /* Add a contents piece to a stream object: size bytes just written on */
 /* streams.strm. */
 int
-cos_stream_add(cos_stream_t *pcs, uint size)
+cos_stream_add(gx_device_pdf *pdev, cos_stream_t *pcs, uint size)
 {
-    gx_device_pdf *pdev = pcs->pdev;
-    stream *s = pdev->streams.strm;
-    gs_offset_t position = stell(s);
+    stream *s;
+    gs_offset_t position;
     cos_stream_piece_t *prev = pcs->pieces;
 
+    /* Beware of subclassed devices. This is mad IMO, we should store
+     * 's' in the stream state somewhere.
+     */
+    while (pdev->child)
+        pdev = (gx_device_pdf *)pdev->child;
+
+    s = pdev->streams.strm;
+    position = stell(s);
+
     /* Check for consecutive writing -- just an optimization. */
     if (prev != 0 && prev->position + prev->size + size == position) {
         prev->size += size;
@@ -1916,15 +1919,15 @@ cos_stream_add(cos_stream_t *pcs, uint size)
 
 /* Add bytes to a stream object. */
 int
-cos_stream_add_bytes(cos_stream_t *pcs, const byte *data, uint size)
+cos_stream_add_bytes(gx_device_pdf *pdev, cos_stream_t *pcs, const byte *data, uint size)
 {
-    stream_write(pcs->pdev->streams.strm, data, size);
-    return cos_stream_add(pcs, size);
+    stream_write(pdev->streams.strm, data, size);
+    return cos_stream_add(pdev, pcs, size);
 }
 
 /* Add the contents of a stream to a stream object. */
 int
-cos_stream_add_stream_contents(cos_stream_t *pcs, stream *s)
+cos_stream_add_stream_contents(gx_device_pdf *pdev, cos_stream_t *pcs, stream *s)
 {
     int code = 0;
     byte sbuff[200];	/* arbitrary */
@@ -1941,19 +1944,17 @@ cos_stream_add_stream_contents(cos_stream_t *pcs, stream *s)
                 break;
             return_error(gs_error_ioerror);
         }
-    } while ((code = cos_stream_add_bytes(pcs, sbuff, cnt)) >= 0);
+    } while ((code = cos_stream_add_bytes(pdev, pcs, sbuff, cnt)) >= 0);
     return code;
 }
 
 /* Release the last contents piece of a stream object. */
 /* Warning : this function can't release pieces if another stream is written after them. */
 int
-cos_stream_release_pieces(cos_stream_t *pcs)
+cos_stream_release_pieces(gx_device_pdf *pdev, cos_stream_t *pcs)
 {
-    gx_device_pdf *pdev = pcs->pdev;
     stream *s = pdev->streams.strm;
     gs_offset_t position = stell(s), position0 = position;
-    gs_memory_t *mem = cos_object_memory((cos_object_t *)pcs);
 
     while (pcs->pieces != NULL &&
                 position == pcs->pieces->position + pcs->pieces->size) {
@@ -1961,7 +1962,7 @@ cos_stream_release_pieces(cos_stream_t *pcs)
 
         position -= p->size;
         pcs->pieces = p->next;
-        gs_free_object(mem, p, "cos_stream_release_pieces");
+        gs_free_object(cos_object_memory((cos_object_t *)pcs), p, "cos_stream_release_pieces");
     }
     if (position0 != position)
         if (sseek(s, position) < 0)
@@ -1990,16 +1991,20 @@ cos_write_stream_process(stream_state * st, stream_cursor_read * pr,
 {
     uint count = pr->limit - pr->ptr;
     cos_write_stream_state_t *ss = (cos_write_stream_state_t *)st;
-    gx_device_pdf *pdev = ss->pdev;
     stream *target = ss->target;
-    gs_offset_t start_pos = stell(pdev->streams.strm);
+    gx_device_pdf *pdev = ss->pdev;
+    gs_offset_t start_pos;
     int code;
 
+    while(pdev->child != NULL)
+        pdev = (gx_device_pdf *)pdev->child;
+
+    start_pos = stell(pdev->streams.strm);
     stream_write(target, pr->ptr + 1, count);
     gs_md5_append(&ss->pcs->md5, pr->ptr + 1, count);
     pr->ptr = pr->limit;
     sflush(target);
-    code = cos_stream_add(ss->pcs, (uint)(stell(pdev->streams.strm) - start_pos));
+    code = cos_stream_add(pdev, ss->pcs, (uint)(stell(pdev->streams.strm) - start_pos));
     return (code < 0 ? ERRC : 0);
 }
 static int
@@ -2007,9 +2012,13 @@ cos_write_stream_close(stream *s)
 {
     cos_write_stream_state_t *ss = (cos_write_stream_state_t *)s->state;
     int status;
+    gx_device_pdf *target_dev = (gx_device_pdf *)ss->pdev;
+
+    while (target_dev->child != NULL)
+        target_dev = (gx_device_pdf *)target_dev->child;
 
     sflush(s);
-    status = s_close_filters(&ss->target, ss->pdev->streams.strm);
+    status = s_close_filters(&ss->target, target_dev->streams.strm);
     gs_md5_finish(&ss->pcs->md5, (gs_md5_byte_t *)ss->pcs->stream_hash);
     ss->pcs->stream_md5_valid = 1;
     return (status < 0 ? status : s_std_close(s));
@@ -2041,6 +2050,8 @@ cos_write_stream_alloc(cos_stream_t *pcs, gx_device_pdf *pdev,
     gs_md5_init(&ss->pcs->md5);
     memset(&ss->pcs->hash, 0x00, 16);
     ss->pdev = pdev;
+    while(ss->pdev->parent)
+        ss->pdev = (gx_device_pdf *)ss->pdev->parent;
     ss->s = s;
     ss->target = pdev->streams.strm; /* not s->strm */
     s_std_init(s, buf, CWS_BUF_SIZE, &cos_s_procs, s_mode_write);
@@ -2060,17 +2071,12 @@ cos_stream_from_pipeline(stream *s)
 {
     cos_write_stream_state_t *ss;
 
-    while(s->procs.process != cos_s_procs.process)
+    while(s->procs.process != cos_s_procs.process) {
         s = s->strm;
+        if (s == 0L)
+            return 0L;
+    }
+
     ss = (cos_write_stream_state_t *)s->state;
     return ss->pcs;
 }
-
-/* Get cos write stream from pipeline. */
-stream *
-cos_write_stream_from_pipeline(stream *s)
-{
-    while(s->procs.process != cos_s_procs.process)
-        s = s->strm;
-    return s;
-}
