@@ -1,4 +1,4 @@
-/* Copyright (C) 2020-2021 Artifex Software, Inc.
+/* Copyright (C) 2020-2022 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -52,7 +52,7 @@ static int pdfi_repair_add_object(pdf_context *ctx, uint64_t obj, uint64_t gen,
         ctx->xref_table->type = PDF_XREF_TABLE;
         ctx->xref_table->xref_size = obj + 1;
 #if REFCNT_DEBUG
-        ctx->xref_table->UID = ctx->UID++;
+        ctx->xref_table->UID = ctx->ref_UID++;
         dmprintf1(ctx->memory, "Allocated xref table with UID %"PRIi64"\n", ctx->xref_table->UID);
 #endif
         pdfi_countup(ctx->xref_table);
@@ -109,24 +109,24 @@ int pdfi_repair_file(pdf_context *ctx)
      */
     pdfi_seek(ctx, ctx->main_stream, 0, SEEK_SET);
     {
-        char Buffer[10], test[] = "%PDF";
+        static const char test[] = "%PDF";
         int index = 0;
 
         do {
-            code = pdfi_read_bytes(ctx, (byte *)&Buffer[index], 1, 1, ctx->main_stream);
-            if (code < 0)
+            int c = pdfi_read_byte(ctx, ctx->main_stream);
+            if (c < 0)
                 goto exit;
 
-            if (Buffer[index] == test[index])
+            if (c == test[index])
                 index++;
             else
                 index = 0;
-        } while (index < 4 && ctx->main_stream->eof == false);
-        if (memcmp(Buffer, test, 4) != 0) {
+        } while (index < 4);
+        if (index != 4) {
             code = gs_note_error(gs_error_undefined);
             goto exit;
         }
-        pdfi_unread(ctx, ctx->main_stream, (byte *)Buffer, 4);
+        pdfi_unread(ctx, ctx->main_stream, (byte *)test, 4);
         pdfi_skip_comment(ctx, ctx->main_stream);
     }
     if (ctx->main_stream->eof == true) {
@@ -188,7 +188,7 @@ int pdfi_repair_file(pdf_context *ctx)
                             /* move all the saved offsets up by one */
                             saved_offset[0] = saved_offset[1];
                             saved_offset[1] = saved_offset[2];
-                            saved_offset[2] = pdfi_unread_tell(ctx);;
+                            saved_offset[2] = pdfi_unread_tell(ctx);
 
                             code = pdfi_read_token(ctx, ctx->main_stream, 0, 0);
                             if (code < 0) {
@@ -228,21 +228,22 @@ int pdfi_repair_file(pdf_context *ctx)
                                     break;
                                 } else {
                                     if (k->key == TOKEN_STREAM) {
-                                        char Buffer[10], test[] = "endstream";
+                                        static const char test[] = "endstream";
                                         int index = 0;
 
                                         do {
-                                            code = pdfi_read_bytes(ctx, (byte *)&Buffer[index], 1, 1, ctx->main_stream);
-                                            if (code < 0) {
-                                                if (code != gs_error_VMerror && code != gs_error_ioerror)
-                                                    continue;
+                                            int c = pdfi_read_byte(ctx, ctx->main_stream);
+                                            if (c == EOFC)
+                                                break;
+                                            if (c < 0)
                                                 goto exit;
-                                            }
-                                            if (Buffer[index] == test[index])
+                                            if (c == test[index])
                                                 index++;
+                                            else if (c == test[0]) /* Pesky 'e' appears twice */
+                                                index = 1;
                                             else
                                                 index = 0;
-                                        } while (index < 9 && ctx->main_stream->eof == false);
+                                        } while (index < 9);
                                         do {
                                             code = pdfi_read_token(ctx, ctx->main_stream, 0, 0);
                                             if (code < 0) {
@@ -250,16 +251,18 @@ int pdfi_repair_file(pdf_context *ctx)
                                                     continue;
                                                 goto exit;
                                             }
-                                            if (ctx->stack_top[-1]->type == PDF_KEYWORD){
-                                                pdf_keyword *k = (pdf_keyword *)ctx->stack_top[-1];
-                                                if (k->key == TOKEN_ENDOBJ) {
-                                                    code = pdfi_repair_add_object(ctx, object_num, generation_num, offset);
-                                                    if (code < 0) {
-                                                        if (code != gs_error_VMerror && code != gs_error_ioerror)
-                                                            break;
-                                                        goto exit;
+                                            if (code > 0) {
+                                                if (ctx->stack_top[-1]->type == PDF_KEYWORD){
+                                                    pdf_keyword *k = (pdf_keyword *)ctx->stack_top[-1];
+                                                    if (k->key == TOKEN_ENDOBJ) {
+                                                        code = pdfi_repair_add_object(ctx, object_num, generation_num, offset);
+                                                        if (code < 0) {
+                                                            if (code != gs_error_VMerror && code != gs_error_ioerror)
+                                                                break;
+                                                            goto exit;
+                                                        }
+                                                        break;
                                                     }
-                                                    break;
                                                 }
                                             }
                                         }while(ctx->main_stream->eof == false);
@@ -287,7 +290,7 @@ int pdfi_repair_file(pdf_context *ctx)
                             } else {
                                 if (k->key == TOKEN_TRAILER) {
                                     code = pdfi_read_bare_object(ctx, ctx->main_stream, 0, 0, 0);
-                                    if (code == 0 && ctx->stack_top[-1]->type == PDF_DICT) {
+                                    if (code == 0 && pdfi_count_stack(ctx) > 0 && ctx->stack_top[-1]->type == PDF_DICT) {
                                         if (ctx->Trailer) {
                                             pdf_dict *d = (pdf_dict *)ctx->stack_top[-1];
                                             bool known = false;
@@ -420,13 +423,15 @@ int pdfi_repair_file(pdf_context *ctx)
                                     if (code == 0) {
                                         for (j=0;j < N; j++) {
                                             code = pdfi_read_token(ctx, compressed_stream, 0, 0);
-                                            if (code == 0) {
+                                            if (code == 0)
+                                                break;
+                                            if (code > 0) {
                                                 o = ctx->stack_top[-1];
                                                 if (((pdf_obj *)o)->type == PDF_INT) {
                                                     obj_num = ((pdf_num *)o)->value.i;
                                                     pdfi_pop(ctx, 1);
                                                     code = pdfi_read_token(ctx, compressed_stream, 0, 0);
-                                                    if (code == 0) {
+                                                    if (code > 0) {
                                                         o = ctx->stack_top[-1];
                                                         if (((pdf_obj *)o)->type == PDF_INT) {
                                                             offset = ((pdf_num *)o)->value.i;
@@ -436,11 +441,16 @@ int pdfi_repair_file(pdf_context *ctx)
                                                                 code = gs_note_error(gs_error_rangecheck);
                                                                 goto exit;
                                                             }
-                                                            ctx->xref_table->xref[obj_num].compressed = true;
-                                                            ctx->xref_table->xref[obj_num].free = false;
-                                                            ctx->xref_table->xref[obj_num].object_num = obj_num;
-                                                            ctx->xref_table->xref[obj_num].u.compressed.compressed_stream_num = i;
-                                                            ctx->xref_table->xref[obj_num].u.compressed.object_index = j;
+                                                            if (obj_num >= ctx->xref_table->xref_size)
+                                                                code = pdfi_repair_add_object(ctx, obj_num, 0, 0);
+
+                                                            if (code >= 0) {
+                                                                ctx->xref_table->xref[obj_num].compressed = true;
+                                                                ctx->xref_table->xref[obj_num].free = false;
+                                                                ctx->xref_table->xref[obj_num].object_num = obj_num;
+                                                                ctx->xref_table->xref[obj_num].u.compressed.compressed_stream_num = i;
+                                                                ctx->xref_table->xref[obj_num].u.compressed.object_index = j;
+                                                            }
                                                         }
                                                     }
                                                 }
